
import requests
import yaml
import os
from tqdm import tqdm
from pydantic import BaseModel, Field
from typing import Callable, Optional
import typing
from inspect import signature
from typing import Any, Callable, Dict, Optional
from bioimageio_chatbot.jsonschema_pydantic import json_schema_to_pydantic_model
from schema_agents import schema_tool

def get_manifest():
    # If no manifest is provided, download from the repo
    if not os.path.exists("./knowledge-base-manifest.yaml"):
        print("Downloading the knowledge base manifest...")
        response = requests.get("https://raw.githubusercontent.com/bioimage-io/bioimageio-chatbot/main/knowledge-base-manifest.yaml")
        assert response.status_code == 200
        with open("./knowledge-base-manifest.yaml", "wb") as f:
            f.write(response.content)
    
    return yaml.load(open("./knowledge-base-manifest.yaml", "r"), Loader=yaml.FullLoader)


def download_file(url, filename):
    response = requests.get(url, stream=True)
    file_size = int(response.headers.get('content-length', 0))

    # Initialize the progress bar
    progress = tqdm(response.iter_content(1024), f"Downloading {filename}", total=file_size, unit="B", unit_scale=True, unit_divisor=1024)
    
    with open(filename, 'wb') as f:
        for data in progress:
            # Update the progress bar
            progress.update(len(data))
            f.write(data)


def extract_schemas(function):
    sig = signature(function)
    positional_annotation = [
        p.annotation
        for p in sig.parameters.values()
        if p.kind == p.POSITIONAL_OR_KEYWORD
    ][0]
    output_schemas = (
        [sig.return_annotation]
        if not isinstance(sig.return_annotation, typing._UnionGenericAlias)
        else list(sig.return_annotation.__args__)
    )
    input_schemas = (
        [positional_annotation]
        if not isinstance(positional_annotation, typing._UnionGenericAlias)
        else list(positional_annotation.__args__)
    )
    return input_schemas, output_schemas

class ChatbotExtension(BaseModel):
    """Chatbot extension."""

    id: str
    name: str
    description: str
    tools: Optional[Dict[str, Any]] = {}
    get_schema: Optional[Callable] = None

class LegacyChatbotExtension(BaseModel):
    """A class that defines the interface for a user extension"""
    name: str = Field(..., description="The name of the extension")
    description: str = Field(..., description="A description of the extension")
    get_schema: Optional[Callable] = Field(None, description="A function that returns the schema for the extension")
    execute: Callable = Field(..., description="The extension's execution function")
    schema_class: Optional[BaseModel] = Field(None, description="The schema class for the extension")

def convert_to_dict(obj):
    if isinstance(obj, BaseModel):
        return obj.dict()
    if isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [convert_to_dict(v) for v in obj]
    return obj


async def legacy_extension_to_tool(extension: LegacyChatbotExtension):
    if extension.get_schema:
        schema = await extension.get_schema()
        extension.schema_class = json_schema_to_pydantic_model(schema)
    else:
        input_schemas, _ = extract_schemas(extension.execute)
        extension.schema_class = input_schemas[0]

    assert extension.schema_class, f"Extension {extension.name} has no valid schema class."

    # NOTE: Right now, the first arguments has to be req
    async def execute(req: extension.schema_class):
        print("Executing extension:", extension.name, req)
        # req = extension.schema_class.parse_obj(req)
        result = await extension.execute(req)
        return convert_to_dict(result)

    execute.__name__ = extension.name

    if extension.get_schema:
        execute.__doc__ = schema['description']
    
    if not execute.__doc__:
        # if extension.execute is partial
        if hasattr(extension.execute, "func"):
            execute.__doc__ = extension.execute.func.__doc__ or extension.description
        else:
            execute.__doc__ = extension.execute.__doc__ or extension.description
    return schema_tool(execute)


import os
from bioimageio_chatbot.chatbot import create_assistants, get_builtin_extensions, QuestionWithHistory, UserProfile
from schema_agents.schema import Message
import pytest

KNOWLEDGE_BASE_PATH = "./bioimageio-knowledge-base"

@pytest.fixture
def builtin_extensions():
    return get_builtin_extensions()

@pytest.fixture
def melman(builtin_extensions):
    assistants = create_assistants(builtin_extensions)
    # find an assistant name Melman
    m = [assistant for assistant in assistants if assistant['name'] == "Melman"][0]
    return m['agent']

@pytest.mark.asyncio
async def test_chatbot(builtin_extensions, melman):
    select_extensions = [
        {"id": "bioimage_archive"}
    ]
    chat_history=[]
    question = "Which tool can I use to analyse western blot image?"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BioimageArchiveSearch" in element for element in str_resp])

    question = "Which tool can I use to segment an cell image?"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BioimageArchiveSearch" in element for element in str_resp])
    
    question = "How can I test the models?"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BioimageArchiveSearch" in element for element in str_resp])

    question = "What are Model Contribution Guidelines?"
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BioimageArchiveSearch" in element for element in str_resp])

    
    # test biii extension
    select_extensions = [
        {"id": "biii"}
    ]
    question = "What bioimage analysis tools are available for quantifying cell migration?"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BiiiSearch" in element for element in str_resp])
    
    question = "Are there any workflows on biii.eu for 3D reconstruction of neuronal networks from electron microscopy images?"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["BiiiSearch" in element for element in str_resp])
    
    
    # test image_sc extension
    select_extensions = [
        {"id": "image_sc_forum"}
    ]
    question = "I got a problem, StarDist stops working! help me find it in image.sc forum."
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    # make resp a string
    resp = [str(element) for element in resp]
    assert any(["ImageScForumSearch" in element for element in resp])
    assert any(['''posts":''' in element for element in resp])


    # test web extension
    select_extensions = [
        {"id": "web"}
    ]
    question = "I want to know more about the BioImage Archive"
    profile = UserProfile(name="lulu", occupation="data scientist", background="machine learning and AI")
    m = QuestionWithHistory(question=question, chat_history=chat_history, user_profile=UserProfile.model_validate(profile), channel_id=None, chatbot_extensions=select_extensions)
    resp = await melman.handle(Message(content="", data=m , role="User"))
    assert resp
    str_resp = [str(element) for element in resp]
    assert any(["WebSearch" in element for element in str_resp])
    assert any(['''"content": ''' in element for element in str_resp])
    


import asyncio
import os
import json
import datetime
import secrets
import aiofiles
from functools import partial
from imjoy_rpc.hypha import login, connect_to_server

from pydantic import BaseModel, Field
from schema_agents import Role, Message
from typing import Any, Dict, List, Optional
import pkg_resources
from bioimageio_chatbot.chatbot_extensions import (
    convert_to_dict,
    get_builtin_extensions,
    extension_to_tools,
    create_tool_name,
)
from bioimageio_chatbot.utils import ChatbotExtension, LegacyChatbotExtension, legacy_extension_to_tool
from bioimageio_chatbot.gpts_action import serve_actions
from bioimageio_chatbot.quota import QuotaManager
import logging


logger = logging.getLogger("bioimageio-chatbot")
# set logger level
logger.setLevel(logging.INFO)


class UserProfile(BaseModel):
    """The user's profile. This will be used to personalize the response to the user."""

    name: str = Field(description="The user's name.", max_length=32)
    occupation: str = Field(description="The user's occupation.", max_length=128)
    background: str = Field(description="The user's background.", max_length=256)


class QuestionWithHistory(BaseModel):
    """The user's question, chat history, and user's profile."""

    question: str = Field(description="The user's question.")
    chat_history: Optional[List[Dict[str, str]]] = Field(
        None, description="The chat history."
    )
    user_profile: Optional[UserProfile] = Field(
        None,
        description="The user's profile. You should use this to personalize the response based on the user's background and occupation.",
    )
    chatbot_extensions: Optional[List[Dict[str, Any]]] = Field(
        None, description="Chatbot extensions."
    )
    context: Optional[Dict[str, Any]] = Field(
        None, description="The context of request."
    )


class ResponseStep(BaseModel):
    """Response step"""

    name: str = Field(description="Step name")
    details: Optional[dict] = Field(None, description="Step details")


class RichResponse(BaseModel):
    """Rich response with text and intermediate steps"""

    text: str = Field(description="Response text")
    steps: List[ResponseStep] = Field(description="Intermediate steps")
    remaining_quota: Optional[float] = Field(None, description="Remaining quota")


def create_assistants(builtin_extensions):
    # debug = os.environ.get("BIOIMAGEIO_DEBUG") == "true"

    async def respond_to_user(
        question_with_history: QuestionWithHistory = None, role: Role = None
    ) -> RichResponse:
        """Response to the user's query."""
        steps = []
        inputs = (
            [question_with_history.user_profile]
            + list(question_with_history.chat_history)
            + [question_with_history.question]
        )
        assert question_with_history.chatbot_extensions is not None
        extensions_by_id = {ext.id: ext for ext in builtin_extensions}
        extensions_by_name = {ext.name: ext for ext in builtin_extensions}
        extensions_by_tool_name = {}
        
        tools = []
        tool_prompts = {}
        for ext in question_with_history.chatbot_extensions:
            if "id" in ext and ext["id"] in extensions_by_id:
                extension = extensions_by_id[ext["id"]]
            elif "name" in ext and ext["name"] in extensions_by_name:
                extension = extensions_by_name[ext["name"]]
            else:
                if "tools" not in ext and "execute" in ext and "get_schema" in ext:
                    # legacy chatbot extension
                    extension = LegacyChatbotExtension.model_validate(ext)
                    logger.warning(f"Legacy chatbot extension is deprecated. Please use the new ChatbotExtension interface for {extension.name} with multi-tool support.")
                else:
                    extension = ChatbotExtension.model_validate(ext)

            max_length = 4000
            if isinstance(extension, LegacyChatbotExtension):
                ts = [await legacy_extension_to_tool(extension)]
                assert len(extension.description) <= max_length, f"Extension description is too long: {extension.description}"
                tool_prompts[create_tool_name(extension.name)] = extension.description.replace("\n", ";")[:max_length]
            else:
                ts = await extension_to_tools(extension)
                assert len(extension.description) <= max_length, f"Extension tool prompt is too long: {extension.description}"
                tool_prompts[create_tool_name(extension.id) + "*"] = extension.description.replace("\n", ";")[:max_length]
            extensions_by_tool_name.update({t.__name__: extension for t in ts})
            tools += ts
            

        class ThoughtsSchema(BaseModel):
            """Details about the thoughts"""

            reasoning: str = Field(
                ...,
                description="reasoning and constructive self-criticism; make it short and concise in less than 20 words",
            )
            # reasoning: str = Field(..., description="brief explanation about the reasoning")
            # criticism: str = Field(..., description="constructive self-criticism")

        tool_usage_prompt = "Tool usage guidelines (* represent the prefix of a tool group):\n" + "\n".join([f" - {ext}:{tool_prompt}" for ext, tool_prompt in tool_prompts.items()])
        response, metadata = await role.acall(
            inputs,
            tools,
            return_metadata=True,
            thoughts_schema=ThoughtsSchema,
            max_loop_count=20,
            tool_usage_prompt=tool_usage_prompt,
        )
        result_steps = metadata["steps"]
        for idx, step_list in enumerate(result_steps):
            steps.append(
                ResponseStep(
                    name=f"step-{idx}", details={"details": convert_to_dict(step_list)}
                )
            )
        return RichResponse(text=response, steps=steps)

    melman = Role(
        instructions="You are Melman from Madagascar, a helpful assistant for the bioimaging community. "
        "You ONLY respond to user's queries related to bioimaging. "
        "Your communications should be accurate, concise, and avoid fabricating information, "
        "and if necessary, request additional clarification."
        "Your goal is to deliver an accurate, complete, and transparent response efficiently.",
        actions=[respond_to_user],
        model="gpt-4o",
    )
    event_bus = melman.get_event_bus()
    event_bus.register_default_events()

    bridget_instructions = (
        "As Bridget, your role is to act as an expert in image analysis, guiding users in utilizing image analysis tools and writing analysis code and scripts effectively, help user to analyse their own data. "
        "Communicate accurately, concisely, and logically, refraining from making up information. "
        "When necessary, seek further details to fully understand the user's request. "
        "Your primary objective is to assist users with actual image analysis task by running code in the Code Interpreter.\n"
        "Engage with users to grasp their data, requirements, solicit additional information as needed, use the web search and code interpreter, loading and preprocess user's data into formats that fit the needs of the tools, break down complex task into executable steps, troubleshooting issues, and addressing user's needs as much as you can."
        "NOTE: You are targeting naive users who are not familiar with programming, so unless requested by the user, don't provide code snippets, only concise explanations and guidance."
    )

    nina_instructions = (
        "As Nina, your focus is on serving as a professional trainer specialized in bioimaging. "
        "Address only inquiries related to bioimaging, ensuring your responses are not only accurate, concise, and logical, but also educational and engaging. "
        "Your mission is to decipher the user's needs through clarifying questions, impart fundamental knowledge of bioimaging, search the associated documentation and books to obtain additional information,"
        "and guide users through the principles and tools of the field, provide concrete examples and suggestions to their question. Offer educational resources, including materials and tutorials, to enhance the user's learning experience."
    )

    bridget = Role(
        instructions=bridget_instructions,
        actions=[respond_to_user],
        model="gpt-4o",
    )

    nina = Role(
        instructions=nina_instructions,
        actions=[respond_to_user],
        model="gpt-4o",
    )
    
    skyler_instructions = (
        "As Skyler, your focus is on serving as an assistant in computational bioimaging. "
        "Address only inquiries related to bioimaging, ensuring your responses are not only accurate, concise, and logical, but also educational and engaging. "
        "Your mission is to decipher the user's needs through clarifying questions, help user by invoking the provided tools."
    )

    skyler = Role(
        instructions=skyler_instructions,
        actions=[respond_to_user],
        model="gpt-4o",
    )


    # convert to a list
    all_extensions = [
        {"id": ext.id, "name": ext.name, "description": ext.description} for ext in builtin_extensions
    ]
    # remove item with 'book' in all_extensions
    melman_extensions = [
        ext for ext in all_extensions if ext["id"] != "books" and ext["id"] != "vision"
    ]
    
    bridget_extensions = [
        ext for ext in all_extensions if ext["id"] == "web" or ext["id"] == "vision"
    ]

    # only keep the item with 'book' in all_extensions
    nina_extensions = [
        ext for ext in all_extensions if "books" == ext["id"]
    ] + [
        ext for ext in all_extensions if ext["id"] == "web"
    ]

    skyler_extensions = []
    return [
        {"name": "Melman", "agent": melman, "extensions": melman_extensions, "code_interpreter": False, "alias": "BioImage Seeker", "icon": "https://bioimage.io/static/img/bioimage-io-icon.svg", "welcome_message": "Hi there! I'm Melman. I am help you navigate the bioimaging tools and provide information about bioimaging. How can I help you today?"},
        {"name": "Nina", "agent": nina, "extensions": nina_extensions, "code_interpreter": False, "alias": "BioImage Tutor", "icon": "https://bioimage.io/static/img/bioimage-io-icon.svg", "welcome_message": "Hi there! I'm Nina, I can help with your learning journey in bioimaging. How can I help you today?"},
        {"name": "Bridget", "agent": bridget, "extensions": bridget_extensions, "code_interpreter": True, "alias": "BioImage Analyst", "icon": "https://bioimage.io/static/img/bioimage-io-icon.svg", "welcome_message": "Hi there! I'm Bridget, I can help you with your bioimaging tasks. Please mount your data folder and let me know how I can assist you today."},
        {"name": "Skyler", "agent": skyler, "extensions": skyler_extensions, "code_interpreter": False, "alias": "BioImage GPT", "icon": "https://bioimage.io/static/img/bioimage-io-icon.svg", "welcome_message": "Hi there! I'm Skyler. How can I help you today?"},
    ]


async def save_chat_history(chat_log_full_path, chat_his_dict):
    # Serialize the chat history to a json string
    chat_history_json = json.dumps(chat_his_dict)

    # Write the serialized chat history to the json file
    async with aiofiles.open(chat_log_full_path, mode="w", encoding="utf-8") as f:
        await f.write(chat_history_json)


async def connect_server(server_url):
    """Connect to the server and register the chat service."""
    login_required = os.environ.get("BIOIMAGEIO_LOGIN_REQUIRED") == "true"
    if login_required:
        token = await login({"server_url": server_url})
    else:
        token = None
    server = await connect_to_server(
        {"server_url": server_url, "token": token, "method_timeout": 100}
    )
    await register_chat_service(server)


async def register_chat_service(server):
    """Hypha startup function."""
    debug = os.environ.get("BIOIMAGEIO_DEBUG") == "true"
    builtin_extensions = get_builtin_extensions()
    login_required = os.environ.get("BIOIMAGEIO_LOGIN_REQUIRED") == "true"
    chat_logs_path = os.environ.get("BIOIMAGEIO_CHAT_LOGS_PATH", "./chat_logs")
    default_quota = float(os.environ.get("BIOIMAGEIO_DEFAULT_QUOTA", "inf"))
    reset_period = os.environ.get("BIOIMAGEIO_DEFAULT_RESET_PERIOD", "hourly")
    quota_database_path = os.environ.get("BIOIMAGEIO_QUOTA_DATABASE_PATH", ':memory:')
    quota_manager = QuotaManager(db_file=quota_database_path, vip_list=[], default_quota=default_quota, default_reset_period=reset_period)
    assert (
        chat_logs_path is not None
    ), "Please set the BIOIMAGEIO_CHAT_LOGS_PATH environment variable to the path of the chat logs folder."
    if not os.path.exists(chat_logs_path):
        print(
            f"The chat session folder is not found at {chat_logs_path}, will create one now."
        )
        os.makedirs(chat_logs_path, exist_ok=True)

    assistants = create_assistants(builtin_extensions)

    def load_authorized_emails():
        if login_required:
            authorized_users_path = os.environ.get("BIOIMAGEIO_AUTHORIZED_USERS_PATH")
            if authorized_users_path:
                assert os.path.exists(
                    authorized_users_path
                ), f"The authorized users file is not found at {authorized_users_path}"
                with open(authorized_users_path, "r") as f:
                    authorized_users = json.load(f)["users"]
                authorized_emails = [
                    user["email"] for user in authorized_users if "email" in user
                ]
            else:
                authorized_emails = None
        else:
            authorized_emails = None
        return authorized_emails

    authorized_emails = load_authorized_emails()

    def check_permission(user):
        if user['is_anonymous']:
            return False
        if authorized_emails is None or user["email"] in authorized_emails:
            return True
        else:
            return False

    async def report(user_report, context=None):
        if login_required and context and context.get("user"):
            assert check_permission(
                context.get("user")
            ), "You don't have permission to report the chat history."
        # get the chatbot version
        version = pkg_resources.get_distribution("bioimageio-chatbot").version
        chat_his_dict = {
            "type": user_report["type"],
            "feedback": user_report["feedback"],
            "conversations": user_report["messages"],
            "session_id": user_report["session_id"],
            "timestamp": str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
            "user": context.get("user"),
            "version": version,
        }
        session_id = user_report["session_id"] + secrets.token_hex(4)
        filename = f"report-{session_id}.json"
        # Create a chat_log.json file inside the session folder
        chat_log_full_path = os.path.join(chat_logs_path, filename)
        await save_chat_history(chat_log_full_path, chat_his_dict)
        print(f"User report saved to {filename}")

    async def talk_to_assistant(
        assistant_name, session_id, user_message: QuestionWithHistory, status_callback, user
    ):
        user = user or {}
        if quota_manager.check_quota(user.get("email")) <= 0:
            raise PermissionError("You have exceeded the quota limit. Please wait for the quota to reset.")
        
        assistant_names = [a["name"] for a in assistants]
        assert (
            assistant_name in assistant_names
        ), f"Assistant {assistant_name} is not found."
        # find assistant by name
        assistant = next(a["agent"] for a in assistants if a["name"] == assistant_name)
        session_id = session_id or secrets.token_hex(8)

        # Listen to the `stream` event
        async def stream_callback(message):
            if message.type in ["function_call", "text"]:
                if message.session.id == session_id:
                    await status_callback(message.model_dump())

        event_bus = assistant.get_event_bus()
        event_bus.on("stream", stream_callback)
        try:
            response = await assistant.handle(
                Message(
                    content="", data=user_message, role="User", session_id=session_id
                )
            )
        except Exception as e:
            event_bus.off("stream", stream_callback)
            raise e
        
        quota_manager.use_quota(user.get("email"), 1.0)
        event_bus.off("stream", stream_callback)
        # get the content of the last response
        response = response[-1].data  # type: RichResponse
        assert isinstance(response, RichResponse)
        response.remaining_quota = quota_manager.check_quota(user.get("email"))
        print(
            f"\nUser: {user_message.question}\nAssistant({assistant_name}): {response.text}\nRemaining quota: {response.remaining_quota}\n"
        )

        if session_id:
            user_message.chat_history.append(
                {"role": "user", "content": user_message.question}
            )
            user_message.chat_history.append(
                {"role": "assistant", "content": response.text, "steps": [step.dict() for step in response.steps]}
            )
            version = pkg_resources.get_distribution("bioimageio-chatbot").version
            chat_his_dict = {
                "conversations": user_message.chat_history,
                "timestamp": str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                "user": user_message.context.get("user"),
                "assistant_name": assistant_name,
                "version": version,
            }
            filename = f"chatlogs-{session_id}.json"
            chat_log_full_path = os.path.join(chat_logs_path, filename)
            await save_chat_history(chat_log_full_path, chat_his_dict)
            print(f"Chat history saved to {filename}")
        return response.model_dump()

    async def chat(
        text,
        chat_history,
        user_profile=None,
        status_callback=None,
        session_id=None,
        extensions=None,
        assistant_name="Melman",
        context=None,
    ):
        if login_required and context and context.get("user"):
            logger.info(f"User: {context.get('user')}, Message: {text}")
            assert check_permission(
                context.get("user")
            ), "You don't have permission to use the chatbot, please sign up and wait for approval"

        m = QuestionWithHistory(
            question=text,
            chat_history=chat_history,
            user_profile=UserProfile.model_validate(user_profile),
            chatbot_extensions=extensions,
            context=context,
        )
        return await talk_to_assistant(assistant_name, session_id, m, status_callback, context.get("user"))

    async def ping(context=None):
        if login_required and context and context.get("user"):
            assert check_permission(
                context.get("user")
            ), "You don't have permission to use the chatbot, please sign up and wait for approval"
        return "pong"

    assistant_keys = ["name", "extensions", "alias", "icon", "welcome_message", "code_interpreter"]
    version = pkg_resources.get_distribution("bioimageio-chatbot").version
    hypha_service_info = await server.register_service(
        {
            "name": "BioImage.IO Chatbot",
            "id": "bioimageio-chatbot",
            "config": {"visibility": "public", "require_context": True},
            "version": version,
            "ping": ping,
            "chat": chat,
            "report": report,
            "assistants": {
                a["name"]: {k: a[k] for k in assistant_keys}
                for a in assistants
            },
        }
    )

    server_info = await server.get_connection_info()

    await serve_actions(server, server_info.public_base_url, builtin_extensions)
    server_url = server.config["public_base_url"]

    service_id = hypha_service_info["id"]
    print("=============================\n")
    if server_url.startswith("http://localhost") or server_url.startswith("http://127.0.0.1"):
        print(f"To test the BioImage.IO Chatbot locally, visit: {server_url}/chat")
    print(
        f"\nThe chat client are available publicly at: https://bioimage.io/chat?server={server_url}&service_id={service_id}\n"
        "\n=============================\n"
    )


if __name__ == "__main__":
    server_url = """https://ai.imjoy.io"""
    loop = asyncio.get_event_loop()
    loop.create_task(connect_server(server_url))
    loop.run_forever()



from bioimageio_chatbot.utils import ChatbotExtension
from schema_agents import schema_tool
from pydantic import Field, BaseModel
from typing import Optional, List, Dict, Any
import pandas as pd
from pathlib import Path
import requests
import re
import os
from bioimageio_chatbot.utils import download_file

class HPAClient:
    def __init__(self):
        self._base_url = 'https://www.proteinatlas.org/download/proteinatlas.tsv.zip'
        folder = Path('./data')
        file_path = os.path.join(folder, 'proteinatlas.tsv.zip')
        # firstly check if the data is already downloaded in the /data folder
        if not os.path.exists(file_path):
            os.makedirs(folder, exist_ok=True)
            # download the data
            download_file(self._base_url, file_path)
        # Load and preprocess data at startup
        self.data = pd.read_csv(file_path, delimiter='\t')
        # Convert all textual data to lowercase strings for faster case-insensitive searching
        self.preprocessed_data = self.data.apply(lambda x: x.astype(str).str.lower())
    
    async def search_hpa(self,
        query: str = Field(..., description="Enter gene names, functions, or disease terms to search in the Human Protein Atlas."),
        limitSize: int = Field(10, gt=0, description="Number of returned items per search.")
    ) -> Dict[str, Any]:
        """Search the Human Protein Atlas for proteins based on a query string, return the top search results."""
        query = query.lower()
        
        # Search for the query in the preprocessed data
        query_results = self.preprocessed_data.apply(lambda x: x.str.contains(query)).sum(axis=1)
        query_results = query_results.sort_values(ascending=False)
        query_results = query_results.head(limitSize)

        selected_columns = ['Gene', 'Gene synonym', 'Ensembl', 
                            'Gene description', 'Subcellular location', 'Subcellular main location', 'Subcellular additional location', 
                            'Biological process', 'Molecular function', 'Uniprot', 'Antibody',
                            'Disease involvement', 'Secretome function', 'CCD Protein', 'CCD Transcript',
                            'Evidence', 'Protein class']
        
        info_list = []
        for index in query_results.index:
            items = self.data.loc[index, selected_columns]
            info_list.append(items.to_dict())
        return info_list

    async def read_protein_info(self,
        ensembl: str = Field(..., description="Ensembl ID of the protein.")
    )-> Dict[str, Any]:
        """Get detailed information about a protein from the Human Protein Atlas."""
        json_link = f"https://www.proteinatlas.org/{ensembl}.json"
        response = requests.get(json_link)
        # check if the request was successful
        response.raise_for_status()
        # return the content
        return response.json()



    async def get_cell_image(self,
        gen: str = Field(..., description="Gene name of the protein."),
        ensembl: str = Field(..., description="Ensembl ID of the protein."),
        section: str = Field("subcellular", description="Section of the Human Protein Atlas to search for the protein. Valid options are 'subcellular', 'tissue',")
        ) -> List[str]:
        """Retrieve a list of cell image links from the Human Protein Atlas, where a specific protein is tagged in the green channel. 
        The results should be rendered as a horizatal table of images and create link (format: `[![](http://..._thumb.jpg)](http://....jpg)`) to the full-size image without the '_thumb' suffix."""
        link_name = f"{ensembl}-{gen}"
        http_link = f"https://www.proteinatlas.org/{link_name}/{section}"
        # read the source code of the page
        response = requests.get(http_link)
        # Search for image links, capturing the part after 'src="'
        pattern = r'src="(?P<url>//images\.proteinatlas\.org/.*?_red_green_thumb\.jpg)"'
        image_links = re.findall(pattern, response.text)
        # replace the 'red_green' with 'blue_red_green_yellow' 
        image_links = [link.replace('red_green', 'blue_red_green_yellow') for link in image_links]
        # Remove '_thumb' from each link and print or process them
        final_image_links = []
        for link in image_links:
            final_image_links.append(f"https:{link}")
        return final_image_links


def get_extension():
    hpa_client = HPAClient()
    search_tool = schema_tool(hpa_client.search_hpa)
    read_tool = schema_tool(hpa_client.read_protein_info)
    get_cell_image_tool = schema_tool(hpa_client.get_cell_image)

    return ChatbotExtension(
        id="hpa",
        name="Human Protein Atlas",
        description="Search the Human Protein Atlas to find human protein-related information, including gene expressions, functions, locations, disease associations, and cell images etc.",
        tools=dict(
            search=search_tool,
            read=read_tool,
            get_cell_image=get_cell_image_tool
        )
    )

if __name__ == "__main__":
    import asyncio
    async def main():
        extension = get_extension()
        query = "brain"
        limitSize = 2
        print(await extension.tools["search"](query=query, limitSize=limitSize))
        # test only one image
        # print(await extension.tools["inspect"](images=[ImageInfo(url="https://bioimage.io/static/img/bioimage-io-icon.png", title="BioImage.io Icon")], query="What is this?", context_description="Inspect the BioImage.io icon."))
    # Run the async function
    asyncio.run(main())


import sqlite3
import time

class QuotaManager:
    def __init__(self, db_file=':memory:', vip_list=None, default_quota=1.0, default_reset_period='daily'):
        self.db_file = db_file
        self.conn = sqlite3.connect(self.db_file)
        self.vip_list = vip_list or []
        self.default_quota = default_quota
        self.default_reset_period = self.period_to_seconds(default_reset_period)
        self._setup_database()

    def _setup_database(self):
        cursor = self.conn.cursor()
        cursor.execute('''
            CREATE TABLE IF NOT EXISTS quotas (
                user_id TEXT PRIMARY KEY, 
                quota REAL, 
                last_reset REAL,
                reset_period INTEGER,  -- Reset period in seconds
                max_quota REAL
            )
        ''')
        self.conn.commit()

    def set_user_quota(self, user_id, max_quota, reset_period):
        reset_seconds = self.period_to_seconds(reset_period)
        cursor = self.conn.cursor()
        cursor.execute('''
            INSERT OR REPLACE INTO quotas (user_id, quota, last_reset, reset_period, max_quota)
            VALUES (?, ?, ?, ?, ?)
        ''', (user_id, max_quota, time.time(), reset_seconds, max_quota))
        self.conn.commit()

    def check_quota(self, user_id):
        if user_id in self.vip_list:
            return float('inf')

        cursor = self.conn.cursor()
        cursor.execute("SELECT quota, last_reset, reset_period, max_quota FROM quotas WHERE user_id = ?", (user_id,))
        row = cursor.fetchone()
        if row:
            quota, last_reset, reset_period, max_quota = row
            if time.time() - last_reset >= reset_period:
                quota = max_quota
                last_reset = time.time()
                cursor.execute("UPDATE quotas SET quota = ?, last_reset = ? WHERE user_id = ?",
                               (quota, last_reset, user_id))
                self.conn.commit()
        else:
            quota, last_reset, reset_period, max_quota = self.default_quota, time.time(), self.default_reset_period, self.default_quota
            cursor.execute("INSERT INTO quotas (user_id, quota, last_reset, reset_period, max_quota) VALUES (?, ?, ?, ?, ?)",
                           (user_id, quota, last_reset, reset_period, max_quota))
            self.conn.commit()
        return quota

    def use_quota(self, user_id, amount):
        if self.check_quota(user_id) >= amount:
            cursor = self.conn.cursor()
            cursor.execute("UPDATE quotas SET quota = quota - ? WHERE user_id = ?",
                           (amount, user_id))
            self.conn.commit()
            return True
        else:
            return False

    def reset_quota(self, user_id):
        cursor = self.conn.cursor()
        cursor.execute("UPDATE quotas SET quota = max_quota, last_reset = ? WHERE user_id = ?",
                       (time.time(), user_id))
        self.conn.commit()

    def period_to_seconds(self, period):
        if period == 'monthly':
            return 30 * 86400
        elif period == 'weekly':
            return 7 * 86400
        elif period == 'daily':
            return 86400
        elif period == 'hourly':
            return 3600
        else:
            raise ValueError("Invalid period. Choose from 'daily', 'weekly', 'monthly'.")

if __name__ == '__main__':
    # Testing the functionality with asserts
    quota_manager = QuotaManager(vip_list=['userVIP'], default_quota=2.0, default_reset_period='daily')
    # quota_manager.set_user_quota('user123', 1.0, 'daily')
    quota_manager.set_user_quota('user234', 10.0, 'weekly')
    quota_manager.set_user_quota('user345', 30.0, 'monthly')

    # Assert initial quotas are set correctly
    assert quota_manager.check_quota('userVIP') == float('inf'), "VIP user quota should be infinite"
    assert quota_manager.check_quota('user123') == 2.0, "user123 initial quota should be 2.0"
    assert quota_manager.check_quota('user234') == 10.0, "user234 initial quota should be 10.0"
    assert quota_manager.check_quota('user345') == 30.0, "user345 initial quota should be 30.0"

    # Assert usage and quota management
    assert quota_manager.use_quota('user123', 0.5) == True, "Should allow using 0.5 quota"
    assert quota_manager.check_quota('user123') == 2-0.5, "user123 quota after use should be 2-0.5"
    assert quota_manager.use_quota('user123', 3.6) == False, "Should not allow using 3.6 quota (not enough left)"
    assert quota_manager.use_quota('userVIP', 1000) == True, "VIP should always be allowed to use quota"
    assert quota_manager.check_quota('user234') == 10.0, "user234 should still have full quota"

    # Test manual reset
    quota_manager.reset_quota('user123')
    assert quota_manager.check_quota('user123') == 2.0, "user123 should have full quota after reset"

    # Display final test results
    print("Tests completed successfully.")



from bioimageio_chatbot.utils import ChatbotExtension
from openai import AsyncOpenAI
from schema_agents import schema_tool
import base64
from pydantic import Field, BaseModel
from typing import Optional, List
import httpx
from PIL import Image
from io import BytesIO
import matplotlib.pyplot as plt
# make sure matplotlib is operating headless (no GUI)
plt.switch_backend("agg")

# Function to encode the image
def encode_image(image_path):
  with open(image_path, "rb") as image_file:
    return base64.b64encode(image_file.read()).decode('utf-8')

async def aask(images, messages, max_tokens=1024):
    aclient = AsyncOpenAI()
    user_message = []
    # download the images and save it into a list of PIL image objects
    img_objs = []
    for image in images:
        async with httpx.AsyncClient() as client:
            response = await client.get(image.url)
            response.raise_for_status()
        try:
            img = Image.open(BytesIO(response.content))
        except Exception as e:
            raise ValueError(f"Failed to read image {image.title or ''} from {image.url}. Error: {e}")
        img_objs.append(img)
    
    if len(img_objs) == 1:
        # plot the image with matplotlib
        plt.imshow(img_objs[0])
        if images[0].title:
            plt.title(images[0].title)
        fig = plt.gcf()
    else:
        # plot them in subplots with matplotlib in a row
        fig, ax = plt.subplots(1, len(img_objs), figsize=(15, 5))
        for i, img in enumerate(img_objs):
            ax[i].imshow(img)
            if images[0].title:
                ax[i].set_title(images[i].title)
    # save the plot to a buffer as png format and convert to base64
    buffer = BytesIO()
    fig.tight_layout()
    # if the image size (width or height) is smaller than 512, use the original size and aspect ratio
    # otherwise set the maximun width of the image to n*512 pixels, where n is the number of images; the maximum total width is 1024 pixels
    fig_width = min(1024, len(img_objs)*512, fig.get_figwidth()*fig.dpi)
    # make sure the pixel size (not inches)
    fig.set_size_inches(fig_width/fig.dpi, fig.get_figheight(), forward=True)
    
    # save fig
    fig.savefig(buffer, format="png")
    buffer.seek(0)
    base64_image = base64.b64encode(buffer.read()).decode("utf-8")
    # append the image to the user message
    user_message.append({
        "type": "image_url",
        "image_url": {
            "url": f"data:image/png;base64,{base64_image}"
        }
    })
    
    
    for message in messages:
        assert isinstance(message, str), "Message must be a string."
        user_message.append({"type": "text", "text": message})

    response = await aclient.chat.completions.create(
        model="gpt-4o",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful AI assistant that help user to inspect the provided images visually based on the context, make insightful comments and answer questions about the provided images."
            },
            {
                "role": "user",
                "content": user_message
            }
        ],
        max_tokens=max_tokens,
    )
    return response.choices[0].message.content

class ImageInfo(BaseModel):
    """Image information."""
    url: str=Field(..., description="The URL of the image.")
    title: Optional[str]=Field(None, description="The title of the image.")

@schema_tool
async def inspect_tool(images: List[ImageInfo]=Field(..., description="A list of images to be inspected, each with a http url and title"), query: str=Field(..., description="user query about the image"),  context_description: str=Field(..., description="describe the context for the visual inspection task")) -> str:
    """Inspect an image using GPT4-Vision."""
    # assert image_url.startswith("http"), "Image URL must start with http."
    for image in images:
        assert image.url.startswith("http"), "Image URL must start with http."
    
    response = await aask(images, [context_description, query])
    return response

def get_extension():
    return ChatbotExtension(
        id="vision",
        name="Vision Inspector",
        description="Perform visual inspection on images using GPT4-Vision model, used for describing images and answer image related questions. The images will be plotted using matplotlib and then sent to the GPT4-Vision model for inspection.",
        tools=dict(
            inspect=inspect_tool
        )
    )

if __name__ == "__main__":
    import asyncio
    async def main():
        extension = get_extension()
        print(await extension.tools["inspect"](images=[ImageInfo(url="https://bioimage.io/static/img/bioimage-io-icon.png", title="BioImage.io Icon"), ImageInfo(url="https://bioimage.io/static/img/bioimage-io-logo.png", title="BioImage.io Logo")], query="What are these?", context_description="Inspect the BioImage.io icon and logo."))
        # test only one image
        # print(await extension.tools["inspect"](images=[ImageInfo(url="https://bioimage.io/static/img/bioimage-io-icon.png", title="BioImage.io Icon")], query="What is this?", context_description="Inspect the BioImage.io icon."))
    # Run the async function
    asyncio.run(main())